package ru.dbotthepony.kommons.collect

import ru.dbotthepony.kommons.util.Listenable
import java.util.concurrent.CopyOnWriteArrayList

/**
 * Proxies all operations to [backingSet] along with generating events for [SetListener]s
 */
open class ListenableSet<E>(protected val backingSet: MutableSet<E>) : MutableSet<E> {
	interface SetListener<E> {
		fun onClear()
		fun onValueAdded(element: E)
		fun onValueRemoved(element: E)
	}

	protected open fun onClear() {
		listeners.forEach { it.callback.onClear() }
	}

	protected open fun onValueAdded(element: E) {
		listeners.forEach { it.callback.onValueAdded(element) }
	}

	protected open fun onValueRemoved(element: E) {
		listeners.forEach { it.callback.onValueRemoved(element) }
	}

	protected inner class Listener(val callback: SetListener<E>) : Listenable.L {
		init {
			listeners.add(this)
		}

		override fun remove() {
			listeners.remove(this)
		}
	}

	protected val listeners = CopyOnWriteArrayList<Listener>()

	fun addListener(listener: SetListener<E>): Listenable.L {
		return Listener(listener)
	}

	fun clearListeners() {
		listeners.clear()
	}

	override fun add(element: E): Boolean {
		if (backingSet.add(element)) {
			onValueAdded(element)
			return true
		}

		return false
	}

	override fun remove(element: E): Boolean {
		if (backingSet.remove(element)) {
			onValueRemoved(element)
			return true
		}

		return false
	}

	override fun addAll(elements: Collection<E>): Boolean {
		var any = false
		elements.forEach { any = add(it) || any }
		return any
	}

	override fun clear() {
		if (backingSet.isNotEmpty()) {
			backingSet.clear()
			onClear()
		}
	}

	override fun removeAll(elements: Collection<E>): Boolean {
		var any = false
		elements.forEach { any = remove(it) || any }
		return any
	}

	override fun retainAll(elements: Collection<E>): Boolean {
		var any = false

		val iterator = iterator()

		for (value in iterator) {
			if (value !in elements) {
				any = true
				iterator.remove()
			}
		}

		return any
	}

	override fun contains(element: E): Boolean {
		return element in backingSet
	}

	override fun containsAll(elements: Collection<E>): Boolean {
		return elements.all { it in backingSet }
	}

	override fun isEmpty(): Boolean {
		return backingSet.isEmpty()
	}

	override val size: Int
		get() = backingSet.size

	override fun iterator(): MutableIterator<E> {
		return object : MutableIterator<E> {
			private val parent = backingSet.iterator()
			private var lastElement: Any? = Mark

			override fun hasNext(): Boolean {
				return parent.hasNext()
			}

			override fun next(): E {
				return parent.next().also { lastElement = it }
			}

			override fun remove() {
				parent.remove()
				val lastElement = lastElement

				if (lastElement !== Mark) {
					this.lastElement = Mark
					onValueRemoved(lastElement as E)
				}
			}
		}
	}

	private object Mark
}
